/* iod2.c - I/O Device = 12-bit DAC, Data Translation */

#include <dos.h>
#include "iod.h"

#define min(a,b)    (((a) < (b)) ? (a) : (b))
#define TOP     1
#define BOTTOM  2
#define CPS     800000L

#define RDATA   0x2EC		/* port for in data (read) */
#define WDATA   0x2EC		/* port for out data (write) */
#define CMDREG  0x2ED		/* command register port (write) */
#define STAREG  0x2ED		/* status register port (read) */
#define CRESET  0x00		/* command -- reset DT board */
#define CCLRER  0x01		/* command -- clear error */
#define CRDERR  0x02		/* command -- read error register */
#define CCLOCK  0x03		/* command -- set clock period */
#define CWDAIM  0x08		/* command -- write d/a (immediate) */
#define CDASET  0x09		/* command -- set d/a parameters */
#define CWDATA  0x0A		/* command -- write d/a */
#define CWDATAC 0x2A		/* command -- write d/a (contin) */
#define CWDATACD 0x3A		/* command -- write d/a (contin+DMA) */
#define CTEST   0x0B		/* command -- test */
#define CRADIM  0x0C		/* command -- read a/d (immediate) */
#define CADSET  0x0D		/* command -- set a/d parameters */
#define CRDATA  0x0E		/* command -- read d/a */
#define CRDATAC 0x2E		/* command -- read d/a (contin) */
#define CRDATACD 0x3E		/* command -- read d/a (contin+DMA) */
#define CSTOP   0xF		/* command -- stop the board */

/* DMA constants */
#define MODEDMA  0x0B		/* mode register */
#define MASKDMA  0x0A		/* mask register - channel 1 or 3 */
#define CLBPDMA  0x0C		/* clear byte pointer */
#define PSR3DMA  0x82		/* page select register - chan 3 */
#define BAR3DMA  0x06		/* base address register - chan 3 */
#define BCR3DMA  0x07		/* byte count register - chan 3 */
#define PSR1DMA  0x83		/* page select register - chan 1 */
#define BAR1DMA  0x02		/* base address register - chan 1 */
#define BCR1DMA  0x03		/* byte count register - chan 1 */

#define MODEW1A  0x55		/* write chan 1 autoinitialize */
#define MODER1A  0x59		/* read chan 1 autoinitialize */
#define MODEW1N  0x45		/* write chan 1 no autoinit */
#define MODER1N  0x49		/* read chan 1 no autoinit */
#define MODEW3A  0x57		/* write chan 3 autoinitialize */
#define MODER3A  0x5B		/* read chan 3 autoinitialize */
#define MODEW3N  0x47		/* write chan 3 no autoinit */
#define MODER3N  0x4B		/* read chan 3 no autoinit */

#define BUFFLEN  0x100		/* output buffer length */
#define SEMIBUF  0x80		/* half the buffer length */
#define BUFFBYT  0x200		/* buffer length in bytes */
#define SEMIBYT  0x100		/* half buffer length in bytes */
#define SRCLEN   0x200		/* source of output buffer */
 /* guarantees BUFFLEN within a dma "page" */

#define rdata_rdy     (status & 0x01)
#define wdata_rdy     (!(status & 0x02))
#define dt_ready      (status & 0x04)
#define comp_error    (status & 0x80)

void stop_iod2();
int output_dac2(SIGPTR, long, long);
int input_adc2(SIGPTR, long, long);

int     chkdtinstalled();
int     chkdtrdy();

static int wr_cmd_reg();
static int wr_data();
static int chkdtinstalled();
static int chkdtrdy();
static void dt_stop();
static void dt_clock();
static void dt_input();
static void dt_output();
static void wait1ms();
static void wait10us();
static void caldelay();

static int wfrlc = 50;		/* wait-for-ready loop count */
static int status = 0;
static unsigned int source[SRCLEN];
static unsigned int *outbuff;
static struct memelem {
    unsigned int seg, off, cbytes;
}       meml;
static int dtpresent = 0;
static int dly_cnt = 0;


int
init_iod2()
{
    unsigned int *memcurr;
    long    laddr;
    adbits = 12;
    adzero = 2048;

    if (dly_cnt <= 0)
        caldelay(10);           /* Calibrate 10 microsecond delay */
    if ((dtpresent = chkdtinstalled()) == 0)
	return (1);		/* Error is DT board is not present */

    dt_stop();			/* Make sure the DT is not running */

    /* check to see if first or second half is in segment */
    memcurr = (unsigned int *) source;
    laddr = _FP_OFF(memcurr) + (((long) _FP_SEG(memcurr)) << 4);
    meml.seg = (laddr >> 16) & 0x000F;
    meml.off = laddr & 0xFFFF;
    meml.cbytes = BUFFBYT - 1;
    if ((0x10000 - meml.off) > BUFFBYT) {
	outbuff = (unsigned int *) source;
    } else {
	outbuff = (unsigned int *) (source + BUFFLEN);
	memcurr = (unsigned int *) (source + BUFFLEN);
	laddr = _FP_OFF(memcurr) + (((long) _FP_SEG(memcurr)) << 4);
	meml.seg = (laddr >> 16) & 0x000F;
	meml.off = laddr & 0xFFFF;
	meml.cbytes = BUFFBYT - 1;
    };

    stop_iod = stop_iod2;
    output_dac = output_dac2;
    input_adc = input_adc2;
    return (0);
}

void
stop_iod2()
{
    if (dtpresent)
	dt_stop();
}

int
output_dac2(buf, npts, rate)
SIGPTR  buf;
long    npts, rate;
{
    if (npts <= 0 || dtpresent == 0)
	return (0);
    if (rate <= 20)
	return (1);
    dt_stop();
    dt_clock(rate);
    dt_output(buf, npts);
    return (0);
}

int
input_adc2(buf, npts, rate)
SIGPTR  buf;
long    npts, rate;
{
    if (npts <= 0 || dtpresent == 0)
	return (0);
    if (rate <= 20)
	return (1);
    dt_stop();
    dt_clock(rate);
    dt_input(buf, npts);
    return (0);
}

static void
dt_stop()
{
    int     i;

    /* stop the Data Translation board */
    (void) _outp(CMDREG, CSTOP);
    for (i = 0; i < 100; i++)
        wait10us();
    (void) _inp(RDATA);
    (void) wr_cmd_reg(CCLRER);
}

static void
dt_clock(rate)
long    rate;
{
    long    ticks;

    ticks = CPS / rate;

    if (!chkdtrdy())
	return;
    /* set clock period to num ticks */
    (void) wr_cmd_reg(CCLOCK);
    wait1ms();
    (void) wr_data((int) (ticks & 0xff));
    wait1ms();
    (void) wr_data((int) ((ticks >> 8) & 0xff));
    wait1ms();
    adrate = ticks * CPS;
    return;
}

static void
dt_output(buf, npts)
SIGPTR  buf;
long    npts;
{
    unsigned int *buffcurr;
    long int totdata;
    unsigned int currcount;
    int     currstate, i, np;
    int     addr_lo, addr_hi, addr_sg, nbyt_lo, nbyt_hi;

    if (!chkdtrdy())
	return;
    /* set the DT board for write d/a */
    (void) wr_cmd_reg(CDASET);	/* set d/a parameters command */
    (void) wr_data(dachan);	/* d/a channel number */
    (void) wr_data(0);		/* dummy conversion number */
    (void) wr_data(1);		/* dummy conversion number */

    /* copy first BUFFLEN values into output buffer */
    for (i = 0; i < BUFFLEN; i++)
	outbuff[i] = *buf++;
    buffcurr = outbuff;
    currstate = TOP;
    totdata = npts - BUFFLEN;
    currcount = SEMIBYT;
    addr_sg = meml.seg & 0xff;
    addr_lo = meml.off & 0xff;
    addr_hi = (meml.off >> 8) & 0x00ff;
    nbyt_lo = meml.cbytes & 0xff;
    nbyt_hi = (meml.cbytes >> 8) & 0x00ff;

    /* set the dma controller */
    (void) _outp(MODEDMA, MODER1A);
    (void) _outp(CLBPDMA, 0);
    (void) _outp(PSR1DMA, addr_sg);
    (void) _outp(BAR1DMA, addr_lo);
    (void) _outp(BAR1DMA, addr_hi);
    (void) _outp(BCR1DMA, nbyt_lo);
    (void) _outp(BCR1DMA, nbyt_hi);
    (void) _outp(MASKDMA, 1);

    /* set write d/a (dma+contin) */
    (void) wr_cmd_reg(CWDATACD);
    while (totdata > 0) {
	if (chkevent())
	    break;
	if (currstate == TOP) {
	    while (currcount > (SEMIBYT - 1)) {
		currcount = _inp(BCR1DMA);
		currcount += (_inp(BCR1DMA) << 8);
	    };
	} else {
	    while (currcount < SEMIBYT) {
		currcount = _inp(BCR1DMA);
		currcount += (_inp(BCR1DMA) << 8);
	    };
	};
	np = min(SEMIBUF, totdata);
	for (i = 0; i < np; i++)
	    buffcurr[i] = *buf++;
	if (totdata < SEMIBUF)	/* pad with zeroes if needed */
	    for (; i < SEMIBUF; i++)
		buffcurr[i] = adzero;
	if (currstate == TOP) {
	    buffcurr = outbuff + SEMIBUF;
	    currstate = BOTTOM;
	} else {
	    buffcurr = outbuff;
	    currstate = TOP;
	};
	totdata -= np;
    };

    dt_stop();
}

static void
dt_input(buf, npts)
SIGPTR  buf;
long    npts;
{
    unsigned int *buffcurr;
    long int totdata;
    unsigned int currcount;
    int     currstate, i, np;
    int     addr_lo, addr_hi, addr_sg, nbyt_lo, nbyt_hi;

    if (!chkdtrdy())
	return;

    /* set the DT board for read a/d */
    (void) wr_cmd_reg(CADSET);	/* set a/d parameters command */
    (void) wr_data(adgain);	/* gain code: 0=1x, 1=2x, 2=4x */
    (void) wr_data(adchan);	/* a/d channel number, start */
    (void) wr_data(adchan);	/* a/d channel number, end */
    (void) wr_data(0);		/* dummy conversion number (>2) */
    (void) wr_data(1);		/* dummy conversion number */

    buffcurr = outbuff;
    currstate = TOP;
    totdata = 0;
    currcount = SEMIBYT;
    addr_sg = meml.seg & 0xff;
    addr_lo = meml.off & 0xff;
    addr_hi = (meml.off >> 8) & 0x00ff;
    nbyt_lo = meml.cbytes & 0xff;
    nbyt_hi = (meml.cbytes >> 8) & 0x00ff;

    /* set the dma controller */
    (void) _outp(MODEDMA, MODEW1A);
    (void) _outp(CLBPDMA, 0);
    (void) _outp(PSR1DMA, addr_sg);
    (void) _outp(BAR1DMA, addr_lo);
    (void) _outp(BAR1DMA, addr_hi);
    (void) _outp(BCR1DMA, nbyt_lo);
    (void) _outp(BCR1DMA, nbyt_hi);
    (void) _outp(MASKDMA, 1);

    /* set read a/d (dma+contin) */
    (void) wr_cmd_reg(CRDATACD);
    while (totdata < npts) {
	if (chkevent())
	    break;
	if (currstate == TOP) {
	    while (currcount > (SEMIBYT - 1)) {
		currcount = _inp(BCR1DMA);
		currcount += (_inp(BCR1DMA) << 8);
	    };
	} else {
	    while (currcount < SEMIBYT) {
		currcount = _inp(BCR1DMA);
		currcount += (_inp(BCR1DMA) << 8);
	    };
	};
	np = min(SEMIBUF, (npts - totdata));
	for (i = 0; i < np; i++)
	    *buf++ = buffcurr[i];
	if (currstate == TOP) {
	    buffcurr = outbuff + SEMIBUF;
	    currstate = BOTTOM;
	} else {
	    buffcurr = outbuff;
	    currstate = TOP;
	};
	totdata += np;
    };

    dt_stop();
    adnpts = totdata;		/* return # points to main program */
}

static int
chkdtrdy()
{
    int     j;

    for (j = 0; j < wfrlc; j++) {
	status = _inp(STAREG);
        wait10us();
	if (dt_ready)
	    return (1);
	if (dtpresent == 0)
	    return (0);
    };
    return (0);
}

static int
chkdtinstalled()
{
    unsigned char testval_1, testval_2;
    int     i;

    (void) _outp(CMDREG, CTEST);
    for (i = 0; i < 10; i++)
	wait10us();
    testval_1 = _inp(RDATA);
    wait10us();
    testval_2 = _inp(RDATA);
    if (testval_1 + 1 == testval_2)
	return (1);
    return (0);
}

static int
wr_cmd_reg(data)
int     data;
{
    int     i, sta;

    for (i = 0; i < wfrlc; i++) {
	sta = _inp(STAREG);
	if ((sta & 2) != 0)	/* check wdata busy bit */
	    continue;
	if ((sta & 4) != 0) {	/* check dt ready bit */
	    sta = _inp(STAREG);
	    if ((sta & 4) != 0) {	/* check dt ready bit */
		(void) _outp(CMDREG, (unsigned char) data);
		return (0);	/* normal return */
	    }
	}
    }
    return (1);			/* error return: data not written */
}

static int
wr_data(data)
int     data;
{
    int     i, sta;

    for (i = 0; i < wfrlc; i++) {
	sta = _inp(STAREG);
	if ((sta & 80) != 0)	/* check comp error bit */
	    break;
	if ((sta & 2) != 0)	/* check comp wdata busy bit */
	    continue;
	if ((sta & 4) == 0) {	/* check dt ready bit */
	    (void) _outp(WDATA, (unsigned char) data);
	    return (0);		/* normal return */
	}
    }
    return (1);			/* error return: data not written */
}

/*********************************************************************/

#define TODC	(*((long *) 0x46C))
#define CTPS 	1193180L            /* clock ticks per second */
#define CNTDWN	64536L              /* countdown from ctps */

#pragma loop_opt(off)

static void
wait1ms()
{
    int i;
   
    for(i = 0; i < 100; i++)
        wait10us();
}

static void
wait10us()
{
    int i;
   
    for(i = 0; i < dly_cnt; i++)
        continue;
}

#pragma loop_opt()

static void
caldelay(microsec)
{
    double  uspci;
    long    t;
    int     n, j, k, dc1, dc2;
    static int ilup = 32;
    static int icnt = 4;

    uspci = (1e6 * CNTDWN) / CTPS; /* microsec. per clock incr. */
    k = (int) (uspci / (microsec * ilup));
    for (dly_cnt = 1; dly_cnt < 256; dly_cnt *= icnt) {
	for (t = TODC + 1; TODC < t;)
	    continue;
	t = TODC;
	for (n = 0; TODC == t; n++)
	    for (j = 0; j < ilup; j++)
		wait10us();
	if (n <= k)
	    break;
    }
    if (dly_cnt >= icnt) {
        dc1 = dly_cnt / icnt;
        dc2 = dly_cnt;
        while (dc1 < dc2 - 1) {
            dly_cnt = (dc1 + dc2) / 2;
            for (t = TODC + 1; TODC < t;)
                continue;
            t = TODC;
            for (n = 0; TODC == t; n++)
                for (j = 0; j < ilup; j++)
                    wait10us();
            if (n <= k) {
                dc2 = dly_cnt;
            } else {
                dc1 = dly_cnt;
            }
        }
    }
}
